/*
 *    Example program for comparing normal and interpolating sprite rotation 
 *    routines.
 *
 *    By David Khling 
 * 
 */

#include <stdlib.h>
#include <allegro.h>
#include "rotate_i.h"

volatile long int ms = 0;

/* timer for counting milliseconds */
void timer ()
{
   ms++;
}

RGB_MAP my_rgb_map;
int     red, green, yellow, blue, white, black, grey, cyan;

/* Create rotation background.
 */
void bkg (BITMAP *bmp)
{
   int Ix, Iy;

   clear (bmp);
   for (Iy = 0; Iy < bmp->h; Iy += 8)
      for (Ix = 0; Ix < bmp->w; Ix += 8)
 	 rectfill (bmp, Ix, Iy, Ix+7, Iy+7, ((Ix&8)^(Iy&8)) ? grey : black);
}

/* Create and allocate rotation sprite.
 */
BITMAP *getsprite ()
{
   BITMAP *bmp = create_bitmap (51, 51);
   if (!bmp)
   {
      set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
      allegro_message("Not enough memory??\n");
      exit (1);
   }
   clear (bmp);
   circlefill (bmp, 25, 25, 24, blue);
   circlefill (bmp, 25, 25, 20, red);
   circlefill (bmp, 25, 25, 17, blue);
   line (bmp, 0, 0, 49, 49, yellow);
   line (bmp, 25, 0, 25, 49, cyan);
   line (bmp, 0, 25, 49, 25, cyan);
   line (bmp, 0, 49, 49, 0, yellow);
   circlefill (bmp, 25, 25, 8, 0);
   text_mode (-1);
   textout_centre (bmp, font, "TEST", 25, 22, white);
   text_mode (0);
   return (bmp);
}

int main (void)
{
   int    Ix, Iy,
          start,
          pics,
          mode,
          modeX = 0,
          xres = 320, yres = 200,
          virtx = 0, virty = 0,
          colorDepth = 8,
          trans = 1;
   BITMAP *sprite,
          *buffer;
   fixed angle;

   allegro_init ();
   install_keyboard(); 
   install_timer ();
   install_int (timer, 1);

   for (mode = 0; mode < 7; mode++)
   {
      switch (mode)
      {
         case 0: 
	    xres = 320; yres = 200; 
	    virtx = 0; virty = 0;
	    colorDepth = 8; modeX = 0;
	    trans = 1;
            break;
         case 1: 
	   check_cpu ();
	   colorDepth = 16; break;
         case 2: colorDepth = 24; break;
         case 3: colorDepth = 24; trans = 0; break;
         case 4: colorDepth = 32; trans = 1; break;
         case 5: 
	   #ifdef ALLEGRO_DOS
	    xres = 320; yres = 240; 
	    virtx = 320; virty = 400;
	    colorDepth = 8; modeX = 1; 
	    break;
	   #else
	    #define GFX_MODEX GFX_SAFE
	    return 0;
	   #endif
      }

      set_color_depth (colorDepth);
      if (set_gfx_mode(colorDepth == 8 && modeX ? GFX_MODEX : GFX_AUTODETECT, 
		       xres, yres, virtx, virty) == -1)
      {
	 set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
	 allegro_message("Cannot set video mode %ix%i,%ibpp%s, skipping...\n",
			 xres, yres, colorDepth, modeX ? " (modeX)" : "");
	 continue;
      }

//      /* REMOVE ME! */ scroll_screen (0, 103);

      if (colorDepth == 8)
	 create_rgb_table (rgb_map = &my_rgb_map, default_palette, NULL);

      yellow = makecol (255, 255, 0);
      red = makecol (255, 0, 0);
      green = makecol (0, 255, 0);
      blue = makecol (0, 0, 255);
      cyan = makecol (0, 255, 255);
      black = makecol (0, 0, 0);
      grey = makecol (127, 127, 127);
      white = makecol (255, 255, 255);

      /* create the rotation sprite */
      sprite = getsprite ();
      if (colorDepth != 8 && trans)
	 /* convert all black pixels to transparent pixels */
	 for (Iy = 0; Iy < sprite->h; Iy++)
	    for (Ix = 0; Ix < sprite->w; Ix++)
	       if (getpixel (sprite, Ix, Iy) == 0)
		  putpixel (sprite, Ix, Iy, bitmap_mask_color (sprite));

      /* create screen buffer */
      if (modeX)
  	 /* in modeX I want to test whether drawing onto planar bimaps works,
	  *  that's why the buffer has to be a subbitmap 
	  */
	 buffer = create_sub_bitmap (screen, 0, 240, 103, 103);
      else
	 buffer = create_bitmap (103, 103);

      if (!buffer)
      {
	 set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
	 allegro_message("Buy more RAM!\n");
	 return 1;
      }

      clear_to_color (screen, black);
      textout (screen, font, "rotate_sprite:", 5, 1, green);
      textout (screen, font, "rotate_sprite_ip:", 165, 1, green);
      textout (screen, font, "scaled (x2):", 5, 72, green);
      textout (screen, font, "scaled (x2):", 165, 72, green);
      textprintf_centre (screen, font, 160, 191, red, 
			 "Color depth: %i%s%s", colorDepth,
			 modeX ? ",ModeX" : "",
			 trans ? "" : ",bkg!=pink");

      /* rotate sprite with both routines until key is pressed... */
      for (angle = 0; !keypressed (); angle += itofix(1))
      {
	 start = ms;

	 bkg (buffer); rotate_sprite (buffer, sprite, 5, 5, angle);
	 blit (buffer, screen, 0, 0, 35, 10, 60, 60);
	 bkg (buffer); rotate_sprite_ip (buffer, sprite, 5, 5, angle);
	 blit (buffer, screen, 0, 0, 195, 10, 60, 60);
         bkg (buffer); 
	 rotate_scaled_sprite (buffer, sprite, 1, 1, angle, itofix (2));
	 blit (buffer, screen, 0, 0, 35, 82, 103, 103);
         bkg (buffer); 
	 rotate_scaled_sprite_ip (buffer, sprite, 1, 1, angle, itofix (2));
	 blit (buffer, screen, 0, 0, 195, 82, 103, 103);

	 /* don't run faster than 20 FPS */
	 while (ms-start < 1000/20);
      }
      while (keypressed ()) readkey ();

      /* measure speed of `rotate_sprite' */
      for (angle = 0, pics = 0,start = ms; !keypressed ();  
	   angle += itofix (1), pics++)
      {
	 bkg (buffer); 
	 rotate_sprite (buffer, sprite, 5, 5, angle);
	 blit (buffer, screen, 0, 0, 35, 10, 60, 60);
	 if (!(pics%100) && ms-start > 0)
	    textprintf_centre (screen, font, 160, 191, yellow,
			       "Speed: %5.0f FPS. Press any key...",
			    (float)(pics)*1000./(ms-start));
      } 
      while (keypressed ()) readkey ();

      /* measure speed of `rotate_sprite_ip' */
      for (angle = 0, pics = 0, start = ms; !keypressed (); 
	   angle += itofix (1), pics++)
      {
	 bkg (buffer); 
	 rotate_sprite_ip (buffer, sprite, 5, 5, angle);
	 blit (buffer, screen, 0, 0, 195, 10, 60, 60);
	 if (!(pics%100) && ms-start > 0)
	    textprintf_centre (screen, font, 160, 191, yellow,
			       "Speed: %5.0f FPS. Press any key...",
			    (float)(pics)*1000./(ms-start));
      }
      while (keypressed ()) readkey ();

      destroy_bitmap (sprite);
      destroy_bitmap (buffer);
   } 
   return 0;
}
END_OF_MAIN()


