/*
 *    Interpolating sprite rotation routines.
 *
 *    By David Khling / original rotation code by Shawn Hargreaves.
 * 
 *    Note that those routine use `makecol'. So make sure the `rgb_map' lookup
 *    table is set in case you're using 8bit graphic modes.
 */

#include <allegro.h>
#include "rotate_i.h"

/* helper macro for rotating sprites in different color depths
 */
#define DO_ROTATE_IP(BITS, SCALE, PREPPIXEL, GETPIXEL, PUTPIXEL, \
                       MODEDEP, GET4PIXELS)                      \
{                                                                             \
   line = y+hgap-dy-1;                                                        \
   PREPPIXEL(SCALE);                                                          \
                                                                              \
   for (dx=wgap-1; dx>=0; dx--, f2x += f2xd, f2y += f2yd, addr += SCALE)      \
   {                                                                          \
      xi = f2x>>16; yi = f2y>>16;                                             \
      MODEDEP;                                                                \
                                                                              \
      if ((unsigned)xi >= (unsigned)sprite->w-1 ||                            \
          (unsigned)yi >= (unsigned)sprite->h-1)                              \
         continue;                                                            \
                                                                              \
      xf = f2x&0xFFFF; yf = f2y&0xFFFF;                                       \
      GETPIXEL(BITS); 		                                              \
                                                                              \
      GET4PIXELS;                                                             \
      SPLIT4PIXELS (BITS);                                                    \
      INTERPOLATE;                                                            \
                                                                              \
      pixel = makecol##BITS (r, g, b);                                        \
      PUTPIXEL(BITS);                                                         \
   }                                                                          \
}

/* helper macro for splitting the colors of 4 neighbor pixels into rgb
 * components
 */
#define SPLIT4PIXELS(BITS) \
   r1 = getr##BITS(c1), g1 = getg##BITS(c1), b1 = getb##BITS(c1);          \
   r2 = getr##BITS(c2), g2 = getg##BITS(c2), b2 = getb##BITS(c2);          \
   r3 = getr##BITS(c3), g3 = getg##BITS(c3), b3 = getb##BITS(c3);          \
   r4 = getr##BITS(c4), g4 = getg##BITS(c4), b4 = getb##BITS(c4)          

/* helper macro for interpolating one channel (red, green or blue)
 */
#define INTERPOLATE_CHANNEL(C) \
   C = (yf*((xf*(C##1-C##2-C##3+C##4) >> 16) - C##1+C##3) >> 16) + \
       (xf*(C##2-C##1) >> 16) + C##1;                              \
   C = C < 0 ? 0 : C

/* helper macro for interpolating the colors of 4 pixels to a resulting pixel
 */
#define INTERPOLATE \
   INTERPOLATE_CHANNEL(r); \
   INTERPOLATE_CHANNEL(g); \
   INTERPOLATE_CHANNEL(b)

/* helper macro for preparing direct bitmap acces in `DO_ROTATE_IP' for 
 * linear bitmaps
 */
#define PREPPIXEL_LINEAR(SCALE) \
   addr = bmp_write_line(bmp, line) + x*SCALE; \
   bmp_read_line(bmp, line)

/* helper macro for reading the current background pixel `DO_ROTATE_IP' for
 * linear bitmaps
 */
#define GETPIXEL_LINEAR(BITS) bc = bmp_read##BITS (addr)

/* helper macro for writing the interpolated pixel in `DO_ROTATE_IP' for 
 * linear bitmaps
 */
#define PUTPIXEL_LINEAR(BITS) bmp_write##BITS (addr, pixel)

/* helper marcro for preparing direct bitmap acces in `DO_ROTATE_IP' for modeX
 * bitmaps
 */
#define PREPPIXEL_MODEX(SCALE) \
   addr = ((unsigned long)bmp->line[line]); \
   modeXcount = x-1;

/* helper macro for reading the current background pixel `DO_ROTATE_IP' for
 * modeX bitmaps  !! OPTIMIZE ME !!
 */
#define GETPIXEL_MODEX(BITS) bc = getpixel (bmp, modeXcount, line)

/* helper macro for writing the interpolated pixel in `DO_ROTATE_IP' for 
 * modeX bitmaps
 */
#define PUTPIXEL_MODEX(BITS) \
    outportw(0x3C4, (0x100<<(modeXcount&3))|2); \
    bmp_write##BITS (addr + (modeXcount>>2), pixel) 

/* helper macro for reading 4 neighbor pixels from 8bit sprites 
 */
#define GET4PIXELS8 \
   c1 = (c1 = sprite->line[yi][xi]) ? c1 : bc;               \
   c2 = (c2 = sprite->line[yi][xi+1]) ? c2 : bc;             \
   c3 = (c3 = sprite->line[yi+1][xi]) ? c3 : bc;             \
   c4 = (c4 = sprite->line[yi+1][xi+1]) ? c4 : bc

/* helper macro for reading 4 neighbor pixels from 16bit sprites 
 */
#define GET4PIXELS16 \
   c1 = (c1 = ((unsigned short *)sprite->line[yi])[xi]) == tc ? bc : c1;     \
   c2 = (c2 = ((unsigned short *)sprite->line[yi])[xi+1]) == tc ? bc : c2;   \
   c3 = (c3 = ((unsigned short *)sprite->line[yi+1])[xi]) == tc ? bc : c3;   \
   c4 = (c4 = ((unsigned short *)sprite->line[yi+1])[xi+1]) == tc ? bc : c4

/* helper macro for reading 4 neighbor pixels from 24bit sprites 
 */
#define GET4PIXELS24 \
   bmp_select(sprite);                                                        \
   c1 = (c1=bmp_read24((unsigned long)(sprite->line[yi]+xi*3))) == tc ?       \
         bc : c1;                                                             \
   c2 = (c2=bmp_read24((unsigned long)(sprite->line[yi]+(xi+1)*3))) == tc ?   \
         bc : c2;                                                             \
   c3 = (c3=bmp_read24((unsigned long)(sprite->line[yi+1]+xi*3))) == tc ?     \
         bc : c3;                                                             \
   c4 = (c4=bmp_read24((unsigned long)(sprite->line[yi+1]+(xi+1)*3))) == tc ? \
         bc : c4;                                                             \
   bmp_select(bmp)

/* helper macro for reading 4 neighbor pixels from 32bit sprites 
 */
#define GET4PIXELS32 \
   c1 = (c1 = ((unsigned long *)sprite->line[yi])[xi]) == tc ? bc : c1;     \
   c2 = (c2 = ((unsigned long *)sprite->line[yi])[xi+1]) == tc ? bc : c2;   \
   c3 = (c3 = ((unsigned long *)sprite->line[yi+1])[xi]) == tc ? bc : c3;   \
   c4 = (c4 = ((unsigned long *)sprite->line[yi+1])[xi+1]) == tc ? bc : c4

/* pivot_sprite_ip:
 *  Interpolating analogon to `pivot_sprite'.
 */
void pivot_sprite_ip(BITMAP *bmp, BITMAP *sprite, int x, int y, int cx, int cy, fixed angle)
{
   pivot_scaled_sprite_ip(bmp, sprite, x, y, cx, cy, angle, itofix(1));
}

/* pivot_scaled_sprite_ip:
 *  Interpolating analogon to `pivot_scaled_sprite'.
 */
void pivot_scaled_sprite_ip(BITMAP *bmp, BITMAP *sprite, int x, int y, int cx, int cy, fixed angle, fixed scale)
{
   x -= fixtoi(sprite->w*scale/2);
   y -= fixtoi(sprite->h*scale/2);

   cx -= sprite->w/2;
   cy -= sprite->h/2;

   x -= fixtoi(fmul(cx*fcos(angle) - cy*fsin(angle), scale));
   y -= fixtoi(fmul(cx*fsin(angle) + cy*fcos(angle), scale));

   rotate_scaled_sprite_ip(bmp, sprite, x, y, angle, scale);
}

/* rotate_sprite_ip:
 *  Interpolating analogon to `rotate_sprite_ip'.
 */
void rotate_sprite_ip(BITMAP *bmp, BITMAP *sprite, int x, int y, fixed angle)
{
   rotate_scaled_sprite_ip(bmp, sprite, x, y, angle, itofix(1));
}

/* rotate_scaled_sprite_ip:
 *  Interpolating analogon to `rotate_scaled_sprite_ip'.
 */
void rotate_scaled_sprite_ip(BITMAP *bmp, BITMAP *sprite, int x, int y, fixed angle, fixed scale)
{
   fixed f1x, f1y, f1xd, f1yd;
   fixed f2x, f2y, f2xd, f2yd;
   fixed w, h;
   fixed dir, dir2, dist;
   int x1, y1, x2, y2;
   int dx, dy;
   int wgap = sprite->w;
   int hgap = sprite->h;
   int pixel;
   unsigned long addr;

   int line;
   long xf, yf,                // float part of `x' and `y'
        xi, yi,                // integer part
        c1, c2, c3, c4,        // colors of pixels for interpolation
        r1, g1, b1,            // color components of those pixels
        r2, g2, b2,
        r3, g3, b3,
        r4, g4, b4,
        r, g, b,               // interpolation's resulting color
  	bc,                    // background color for one interplation
        tc;                    // transparent color

   /* rotate the top left pixel of the sprite */
   w = itofix(wgap/2);
   h = itofix(hgap/2);
   dir = fatan2(h, w);
   dir2 = fatan2(h, -w);
   dist = fsqrt((wgap*wgap + hgap*hgap) << 6) << 4;
   f1x = w - fmul(dist, fcos(dir - angle));
   f1y = h - fmul(dist, fsin(dir - angle));

   /* map the destination y axis onto the sprite */
   f1xd = fcos(itofix(64) - angle);
   f1yd = fsin(itofix(64) - angle);

   /* map the destination x axis onto the sprite */
   f2xd = fcos(-angle);
   f2yd = fsin(-angle);

   /* scale the sprite? */
   if (scale != itofix(1)) {
      dist = fmul(dist, scale);
      wgap = fixtoi(fmul(itofix(wgap), scale));
      hgap = fixtoi(fmul(itofix(hgap), scale));
      scale = fdiv(itofix(1), scale);
      f1xd = fmul(f1xd, scale);
      f1yd = fmul(f1yd, scale);
      f2xd = fmul(f2xd, scale);
      f2yd = fmul(f2yd, scale);
   }

   /* adjust the size of the region to be scanned */
   x1 = fixtoi(fmul(dist, fcos(dir + angle)));
   y1 = fixtoi(fmul(dist, fsin(dir + angle)));

   x2 = fixtoi(fmul(dist, fcos(dir2 + angle)));
   y2 = fixtoi(fmul(dist, fsin(dir2 + angle)));

   x1 = MAX(ABS(x1), ABS(x2)) - wgap/2;
   y1 = MAX(ABS(y1), ABS(y2)) - hgap/2;

   x -= x1;
   wgap += x1 * 2;
   f1x -= f2xd * x1;
   f1y -= f2yd * x1;

   y -= y1;
   hgap += y1 * 2;
   f1x -= f1xd * y1;
   f1y -= f1yd * y1;

   /* clip the output rectangle */
   if (bmp->clip) {
      while (x < bmp->cl) {
	 x++;
	 wgap--;
	 f1x += f2xd;
	 f1y += f2yd;
      }

      while (y < bmp->ct) {
	 y++;
	 hgap--;
	 f1x += f1xd;
	 f1y += f1yd;
      }

      while (x+wgap > bmp->cr)
	 wgap--;

      while (y+hgap > bmp->cb)
	 hgap--;

      if ((wgap <= 0) || (hgap <= 0))
	 return;
   }
   
   /* and trace a bunch of lines through the bitmaps */
   bmp_select (bmp);
   tc = bitmap_mask_color (sprite);

   for (dy=hgap-1; dy>=0; dy--) {
      f2x = f1x;
      f2y = f1y;

      if (is_linear_bitmap(bmp)) {           /* draw onto a linear bitmap */
	 switch (bitmap_color_depth(bmp)) {

	    #ifdef ALLEGRO_COLOR8
	       case 8: 
		  DO_ROTATE_IP (8, 1, PREPPIXEL_LINEAR, GETPIXEL_LINEAR, 
			 	PUTPIXEL_LINEAR, ,GET4PIXELS8);
	          break;
	    #endif

	    #ifdef ALLEGRO_COLOR16
	       case 15:
	       case 16: 
		  DO_ROTATE_IP (16, 2, PREPPIXEL_LINEAR, GETPIXEL_LINEAR,
				PUTPIXEL_LINEAR, ,GET4PIXELS16);
                  break;
	    #endif

	    #ifdef ALLEGRO_COLOR24
	       case 24: 
		  DO_ROTATE_IP (24, 3, PREPPIXEL_LINEAR, GETPIXEL_LINEAR, 
			 	PUTPIXEL_LINEAR, ,GET4PIXELS24);
	          break;
	    #endif

	    #ifdef ALLEGRO_COLOR32
	       case 32: 
		  DO_ROTATE_IP (32, 4, PREPPIXEL_LINEAR, GETPIXEL_LINEAR, 
			 	PUTPIXEL_LINEAR, ,GET4PIXELS32);
	          break;
	    #endif
	 }
      }

   #ifdef GFX_MODEX
      else {
	  int modeXcount;

          DO_ROTATE_IP (8, 0, PREPPIXEL_MODEX, GETPIXEL_MODEX, 
                        PUTPIXEL_MODEX, modeXcount++, GET4PIXELS8);
      }
   #endif   /* GFX_MODEX */

      f1x += f1xd;
      f1y += f1yd;
   }

   bmp_unwrite_line (bmp);
}






